/*
 * albe.c - albe potential
 *
 * author: Frank Zirkelbach <frank.zirkelbach@physik.uni-augsburg.de>
 *
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>

#include "../moldyn.h"
#include "../math/math.h"
#include "albe.h"

/* create mixed terms from parameters and set them */
int albe_mult_set_params(t_moldyn *moldyn,int element1,int element2) {

	t_albe_mult_params *p;

	// set cutoff before parameters (actually only necessary for some pots)
	if(moldyn->cutoff==0.0) {
		printf("[albe] WARNING: no cutoff!\n");
		return -1;
	}

	/* alloc mem for potential parameters */
	moldyn->pot_params=malloc(sizeof(t_albe_mult_params));
	if(moldyn->pot_params==NULL) {
		perror("[albe] pot params alloc");
		return -1;
	}

	/* these are now albe parameters */
	p=moldyn->pot_params;

	// only 1 combination by now :p
	switch(element1) {
		case SI:
			/* type: silicon */
			p->S[0]=ALBE_S_SI;
			p->R[0]=ALBE_R_SI;
			p->A[0]=ALBE_A_SI;
			p->B[0]=ALBE_B_SI;
			p->r0[0]=ALBE_R0_SI;
			p->lambda[0]=ALBE_LAMBDA_SI;
			p->mu[0]=ALBE_MU_SI;
			p->gamma[0]=ALBE_GAMMA_SI;
			p->c[0]=ALBE_C_SI;
			p->c2[0]=p->c[0]*p->c[0];
			p->d[0]=ALBE_D_SI;
			p->d2[0]=p->d[0]*p->d[0];
			p->c2d2[0]=p->c2[0]/p->d2[0];
			p->h[0]=ALBE_H_SI;
			switch(element2) {
				case C:
					/* type: carbon */
					p->S[1]=ALBE_S_C;
					p->R[1]=ALBE_R_C;
					p->A[1]=ALBE_A_C;
					p->B[1]=ALBE_B_C;
					p->r0[1]=ALBE_R0_C;
					p->lambda[1]=ALBE_LAMBDA_C;
					p->mu[1]=ALBE_MU_C;
					p->gamma[1]=ALBE_GAMMA_C;
					p->c[1]=ALBE_C_C;
					p->c2[1]=p->c[1]*p->c[1];
					p->d[1]=ALBE_D_C;
					p->d2[1]=p->d[1]*p->d[1];
					p->c2d2[1]=p->c2[1]/p->d2[1];
					p->h[1]=ALBE_H_C;
					/* mixed type: silicon carbide */
					p->Smixed=ALBE_S_SIC;
					p->Rmixed=ALBE_R_SIC;
					p->Amixed=ALBE_A_SIC;
					p->Bmixed=ALBE_B_SIC;
					p->r0_mixed=ALBE_R0_SIC;
					p->lambda_m=ALBE_LAMBDA_SIC;
					p->mu_m=ALBE_MU_SIC;
					p->gamma_m=ALBE_GAMMA_SIC;
					p->c_mixed=ALBE_C_SIC;
					p->c2_mixed=p->c_mixed*p->c_mixed;
					p->d_mixed=ALBE_D_SIC;
					p->d2_mixed=p->d_mixed*p->d_mixed;
					p->c2d2_m=p->c2_mixed/p->d2_mixed;
					p->h_mixed=ALBE_H_SIC;
					break;
				default:
					printf("[albe] WARNING: element2\n");
					return -1;
			}
			break;
		default:
			printf("[albe] WARNING: element1\n");
			return -1;
	}

	printf("[albe] parameter completion\n");
	p->S2[0]=p->S[0]*p->S[0];
	p->S2[1]=p->S[1]*p->S[1];
	p->S2mixed=p->Smixed*p->Smixed;
	p->c2[0]=p->c[0]*p->c[0];
	p->c2[1]=p->c[1]*p->c[1];
	p->c2_mixed=p->c_mixed*p->c_mixed;
	p->d2[0]=p->d[0]*p->d[0];
	p->d2[1]=p->d[1]*p->d[1];
	p->d2_mixed=p->d_mixed*p->d_mixed;
	p->c2d2[0]=p->c2[0]/p->d2[0];
	p->c2d2[1]=p->c2[1]/p->d2[1];
	p->c2d2_m=p->c2_mixed/p->d2_mixed;

	printf("[albe] mult parameter info:\n");
	printf("  S (A)  | %f | %f | %f\n",p->S[0],p->S[1],p->Smixed);
	printf("  R (A)  | %f | %f | %f\n",p->R[0],p->R[1],p->Rmixed);
	printf("  A (eV) | %f | %f | %f\n",p->A[0]/EV,p->A[1]/EV,p->Amixed/EV);
	printf("  B (eV) | %f | %f | %f\n",p->B[0]/EV,p->B[1]/EV,p->Bmixed/EV);
	printf("  lambda | %f | %f | %f\n",p->lambda[0],p->lambda[1],
	                                  p->lambda_m);
	printf("  mu     | %f | %f | %f\n",p->mu[0],p->mu[1],p->mu_m);
	printf("  gamma  | %f | %f | %f\n",p->gamma[0],p->gamma[1],p->gamma_m);
	printf("  c      | %f | %f | %f\n",p->c[0],p->c[1],p->c_mixed);
	printf("  d      | %f | %f | %f\n",p->d[0],p->d[1],p->d_mixed);
	printf("  c2     | %f | %f | %f\n",p->c2[0],p->c2[1],p->c2_mixed);
	printf("  d2     | %f | %f | %f\n",p->d2[0],p->d2[1],p->d2_mixed);
	printf("  c2d2   | %f | %f | %f\n",p->c2d2[0],p->c2d2[1],p->c2d2_m);
	printf("  h      | %f | %f | %f\n",p->h[0],p->h[1],p->h_mixed);

	return 0;
}

/* first i loop */
int albe_mult_i0(t_moldyn *moldyn,t_atom *ai) {

	t_albe_mult_params *params;
	t_albe_exchange *exchange;
	
	int i;

	params=moldyn->pot_params;
	exchange=&(params->exchange);

	/* zero exchange values */
	memset(exchange->zeta,0,ALBE_MAXN*sizeof(double));
	for(i=0;i<ALBE_MAXN;i++)
		memset(exchange->dzeta[i],0,ALBE_MAXN*sizeof(t_3dvec));
	exchange->jcnt=0;
	exchange->j2cnt=0;

	return 0;
}

/* first j loop within first i loop */
int albe_mult_i0_j0(t_moldyn *moldyn,t_atom *ai,t_atom *aj,u8 bc) {

	t_albe_mult_params *params;
	t_albe_exchange *exchange;

	double S2,S,R,d2,d,s_r,arg;
	t_3dvec dist;
	int j;
	u8 brand;
	
	params=moldyn->pot_params;
	exchange=&(params->exchange);

	/* get j counter */
	j=exchange->jcnt;

	/* set ij depending values */
	brand=ai->brand;
	if(brand==aj->brand) {
		S2=params->S2[brand];
	}
	else {
		S2=params->S2mixed;
	}

	/* dist_ij, d_ij2 */
	v3_sub(&dist,&(aj->r),&(ai->r));
	if(bc) check_per_bound(moldyn,&dist);
	exchange->dist[j]=dist;
	d2=v3_absolute_square(&dist);
	exchange->d2[j]=d2;

	/* if d_ij2 > S2 => no force & potential energy contribution */
	if(d2>S2) {
		moldyn->run3bp=0;
		exchange->skip[j]=1;
		exchange->jcnt+=1;
		return 0;
	}
	exchange->skip[j]=0;

	/* more ij depending values */
	if(brand==aj->brand) {
		R=params->R[brand];
		S=params->S[brand];
		/* albe needs i,(j/k) depending c,d,h and gamma values */
		exchange->gamma_[j]=&(params->gamma[brand]);
		exchange->c_[j]=&(params->c[brand]);
		exchange->d_[j]=&(params->d[brand]);
		exchange->h_[j]=&(params->h[brand]);
		exchange->c2_[j]=&(params->c2[brand]);
		exchange->d2_[j]=&(params->d2[brand]);
		exchange->c2d2_[j]=&(params->c2d2[brand]);
	}
	else {
		R=params->Rmixed;
		S=params->Smixed;
		/* albe needs i,(j/k) depending c,d,h and gamma values */
		exchange->gamma_[j]=&(params->gamma_m);
		exchange->c_[j]=&(params->c_mixed);
		exchange->d_[j]=&(params->d_mixed);
		exchange->h_[j]=&(params->h_mixed);
		exchange->c2_[j]=&(params->c2_mixed);
		exchange->d2_[j]=&(params->d2_mixed);
		exchange->c2d2_[j]=&(params->c2d2_m);
	}

	/* d_ij */
	d=sqrt(d2);
	exchange->d[j]=d;
	
	/* f_c, df_c */
	if(d<R) {
		exchange->f_c[j]=1.0;
		exchange->df_c[j]=0.0;
	}
	else {
		s_r=S-R;
		arg=M_PI*(d-R)/s_r;
		exchange->f_c[j]=0.5+0.5*cos(arg);
		exchange->df_c[j]=0.5*sin(arg)*(M_PI/(s_r*d));
	}

	/* reset k counter */
	exchange->kcnt=0;

	return 0;
}

/* first k loop within first j loop within first i loop */
int albe_mult_i0_j0_k0(t_moldyn *moldyn,
                       t_atom *ai,t_atom *aj,t_atom *ak,u8 bc) {

	t_albe_mult_params *params;
	t_albe_exchange *exchange;

	int j,k;
	t_3dvec distj,distk;
	double dj,dk,djdk_inv,cos_theta;
	double gj,dgj,h_cos_j,d2_h_cos2_j,frac_j;
	double gk,dgk,h_cos_k,d2_h_cos2_k,frac_k;
	t_3dvec dcosdrj,dcosdrk,tmp;
	t_3dvec *dzjj,*dzkk,*dzjk,*dzkj;

	params=moldyn->pot_params;
	exchange=&(params->exchange);

	if(aj==ak) {
		exchange->kcnt+=1;
		return 0;
	}

	/* k<j & check whether to run k */
	j=exchange->jcnt;
	k=exchange->kcnt;
	if(k>=ALBE_MAXN) {
		printf("FATAL: too many neighbours! (%d)\n",k);
		printf("  atom i:%d | j:%d | k:%d\n",ai->tag,aj->tag,ak->tag);
	}
	if((k>=j)|(exchange->skip[k])) {
		exchange->kcnt+=1;
		return 0;
	}

	/* distances */
	distj=exchange->dist[j];
	distk=exchange->dist[k];
	dj=exchange->d[j];
	dk=exchange->d[k];
	djdk_inv=1.0/(dj*dk);

	/* cos theta */
	cos_theta=v3_scalar_product(&distj,&distk)*djdk_inv;

	/* g(cos(theta)) ij and ik values */
	h_cos_j=*(exchange->h_[j])+cos_theta; // + in albe formalism
	d2_h_cos2_j=*exchange->d2_[j]+(h_cos_j*h_cos_j);
	frac_j=*exchange->c2_[j]/d2_h_cos2_j;
	gj=1.0+*exchange->c2d2_[j]-frac_j;
	gj*=*(exchange->gamma_[j]);
	dgj=*(exchange->gamma_[j])*2.0*frac_j*h_cos_j/d2_h_cos2_j; // + in albe 
	if(ak->brand==aj->brand) {
		gk=gj;
		dgk=dgj;
	}
	else {
		h_cos_k=*(exchange->h_[k])+cos_theta;
		d2_h_cos2_k=*exchange->d2_[k]+(h_cos_k*h_cos_k);
		frac_k=*exchange->c2_[k]/d2_h_cos2_k;
		gk=1.0+*exchange->c2d2_[k]-frac_k;
		gk*=*(exchange->gamma_[k]);
		dgk=*(exchange->gamma_[k])*2.0*frac_k*h_cos_k/d2_h_cos2_k;
	}

#ifdef DEBUG
	if(ai==&(moldyn->atom[DATOM])) 
		printf("zeta_ij: %f %f %f %f\n",f_c_ik*g,f_c_ik,g,d_ik);
#endif

	/* store even more data for second k loop */
	exchange->g[kcount]=g;
	exchange->dg[kcount]=dg;
	exchange->d_ik[kcount]=d_ik;
	exchange->cos_theta[kcount]=cos_theta;
	exchange->f_c_ik[kcount]=f_c_ik;
	exchange->df_c_ik[kcount]=df_c_ik;

	/* increase k counter */
	exchange->kcnt+=1;
		
	return 0;
}

/* first j loop within first i loop */
int albe_mult_i0_j1(t_moldyn *moldyn,t_atom *ai,t_atom *aj,u8 bc) {

	t_albe_mult_params *params;
	t_albe_exchange *exchange;

	params=moldyn->pot_params;
	exchange=&(params->exchange);

	/* increase j counter */
	exchange->jcnt+=1;

	return 0;
}

/* second j loop within first i loop */
int albe_mult_i0_j2(t_moldyn *moldyn,t_atom *ai,t_atom *aj,u8 bc) {

	t_albe_mult_params *params;
	t_albe_exchange *exchange;

	int j;
	double d,f_a,df_a,f_r,df_r,f_c,df_c,b,db;
	double A,B,mu,lambda,r0;
	double energy;
	t_3dvec *dist,force;
	double scale;
	u8 brand;

	params=moldyn->pot_params;
	exchange=&(params->exchange);

	/* get j counter */
	j=exchange->j2cnt;

	/* skip if j not within cutoff */
	if(exchange->skip[j]) {
		moldyn->run3bp=0;
		exchange->j2cnt+=1;
		return 0;
	}

	/* distance */
	d=exchange->d[j];
	dist=&(exchange->dist[j]);
	f_c=exchange->f_c[j];
	df_c=exchange->df_c[j];

	/* determine parameters to calculate fa, dfa, fr, dfr */
	brand=aj->brand;
	if(brand==ai->brand) {
		B=params->B[brand];
		A=params->A[brand];
		r0=params->r0[brand];
		mu=params->mu[brand];
		lambda=params->lambda[brand];
	}
	else {
		B=params->Bmixed;
		A=params->Amixed;
		r0=params->r0_mixed;
		mu=params->mu_m;
		lambda=params->lambda_m;
	}

	/* f_a, df_a */
	f_a=-B*exp(-mu*(d-r0));
	df_a=mu*f_a/d;

	/* f_r, df_r */
	f_r=A*exp(-lambda*(d-r0));
	df_r=lambda*f_r/d;

	/* b, db */
	b=1.0/sqrt(1.0+exchange->zeta[j]);
	db=-0.5*b/(1.0+exchange->zeta[j]);

	/* energy contribution */
	energy=0.5*f_c*(f_r-b*f_a); // - in albe formalism
	moldyn->energy+=energy;
	ai->e+=energy;

	/* force contribution for atom i due to ij bond */
	scale=-0.5*(f_c*(df_r-b*df_a)+df_c*(f_r-b*f_a)); // - in albe formalism
	v3_scale(&force,dist,scale);
	v3_add(&(ai->f),&(ai->f),&force);

#ifdef NDEBUG
if(ai->tag==0) {
printf("force: %.15f %.15f %.15f | %d %d (ij) %.15f\n",force.x,force.y,force.z,ai->tag,aj->tag,exchange->zeta[j]);
printf("    t: %.15f %.15f %.15f\n",ai->f.x,ai->f.y,ai->f.z);
}
#endif

	/* force contribution for atom j due to ij bond */
	v3_scale(&force,&force,-1.0); // dri rij = - drj rij
	v3_add(&(aj->f),&(aj->f),&force);

	/* virial */
	virial_calc(ai,&force,&(exchange->dist_ij));

#ifdef DEBUG
	if((ai==&(moldyn->atom[DATOM]))|(aj==&(moldyn->atom[DATOM]))) {
		printf("force 3bp (j2): [%d %d sum]\n",ai->tag,aj->tag);
		printf("  adding %f %f %f\n",force.x,force.y,force.z);
		if(ai==&(moldyn->atom[DATOM]))
			printf("  total i: %f %f %f\n",ai->f.x,ai->f.y,ai->f.z);
		if(aj==&(moldyn->atom[DATOM]))
			printf("  total j: %f %f %f\n",aj->f.x,aj->f.y,aj->f.z);
		printf("  energy: %f = %f %f %f %f\n",0.5*f_c*(b*f_a+f_r),
		                                    f_c,b,f_a,f_r);
		printf("          %f %f %f\n",exchange->zeta_ij,.0,.0);
	}
#endif

	/* virial */
	virial_calc(ai,&force,dist);

	/* dzeta prefactor = - f_c f_a db, (* -0.5 due to force calc) */
	exchange->pre_dzeta=0.5*f_a*f_c*db;

	/* force contribution (drj derivative) */
	v3_scale(&force,&(exchange->dzeta[j][j]),exchange->pre_dzeta);
	v3_add(&(aj->f),&(aj->f),&force);

#ifdef NDEBUG
if(aj->tag==0) {
printf("force: %.15f %.15f %.15f | %d %d (j der)\n",force.x,force.y,force.z,aj->tag,ai->tag);
printf("    t: %.15f %.15f %.15f\n",aj->f.x,aj->f.y,aj->f.z);
}
#endif

	/* virial */
	virial_calc(ai,&force,dist);

	v3_scale(&force,&force,-1.0);
	v3_add(&(ai->f),&(ai->f),&force);

#ifdef NDEBUG
if(ai->tag==0) {
printf("force: %.15f %.15f %.15f | %d %d (i contr j der)\n",force.x,force.y,force.z,ai->tag,aj->tag);
printf("    t: %.15f %.15f %.15f\n",ai->f.x,ai->f.y,ai->f.z);
}
#endif

	/* reset k counter for second k loop */
	exchange->kcnt=0;
		
	return 0;
}

/* second k loop within second j loop within first i loop */
int albe_mult_i0_j2_k0(t_moldyn *moldyn,
                       t_atom *ai,t_atom *aj,t_atom *ak,u8 bc) {

	t_albe_mult_params *params;
	t_albe_exchange *exchange;

	int j,k;
	t_3dvec force;

	params=moldyn->pot_params;
	exchange=&(params->exchange);

	if(aj==ak) {
		exchange->kcnt+=1;
		return 0;
	}

	/* prefactor dzeta */
	pre_dzeta=exchange->pre_dzeta;

	/* dist_ik, d_ik */
	dist_ik=exchange->dist_ik[kcount];
	d_ik=exchange->d_ik[kcount];

	/* f_c_ik, df_c_ik */
	f_c_ik=exchange->f_c_ik[kcount];
	df_c_ik=exchange->df_c_ik[kcount];

	/* dist_ij, d_ij2, d_ij */
	dist_ij=exchange->dist_ij;
	d_ij2=exchange->d_ij2;
	d_ij=exchange->d_ij;

	/* g, dg, cos_theta */
	g=exchange->g[kcount];
	dg=exchange->dg[kcount];
	cos_theta=exchange->cos_theta[kcount];

	/* cos_theta derivatives wrt j,k */
	dijdik_inv=1.0/(d_ij*d_ik);
	v3_scale(&dcosdrj,&dist_ik,dijdik_inv);		// j
	v3_scale(&tmp,&dist_ij,-cos_theta/d_ij2);
	v3_add(&dcosdrj,&dcosdrj,&tmp);
	v3_scale(&dcosdrk,&dist_ij,dijdik_inv);		// k
	v3_scale(&tmp,&dist_ik,-cos_theta/d_ik2);
	v3_add(&dcosdrk,&dcosdrk,&tmp);

	/* f_c_ik * dg, df_c_ik * g */
	fcdg=f_c_ik*dg;
	dfcg=df_c_ik*g;

	/* derivative wrt j */
	v3_scale(&force,&dcosdrj,fcdg*pre_dzeta);

	/* force contribution */
	v3_add(&(aj->f),&(aj->f),&force);

#ifdef DEBUG
	if(aj==&(moldyn->atom[DATOM])) {
		printf("force 3bp (k2): [%d %d %d]\n",ai->tag,aj->tag,ak->tag);
		printf("  adding %f %f %f\n",force.x,force.y,force.z);
		printf("  total j: %f %f %f\n",aj->f.x,aj->f.y,aj->f.z);
		printf("  angle: %f\n",acos(cos_theta)*360.0/(2*M_PI));
		printf("    d ij ik = %f %f\n",d_ij,d_ik);
	}
#endif

	/* virial */
	virial_calc(ai,&force,&dist_ij);

	/* force contribution to atom i */
	v3_scale(&force,&force,-1.0);
	v3_add(&(ai->f),&(ai->f),&force);

	/* derivative wrt k */
	v3_scale(&force,&dist_ik,-1.0*dfcg); // dri rik = - drk rik
	v3_scale(&tmp,&dcosdrk,fcdg);
	v3_add(&force,&force,&tmp);
	v3_scale(&force,&force,pre_dzeta);

	v3_scale(&force,&force,-1.0);
	v3_add(&(ai->f),&(ai->f),&force);

#ifdef DEBUG
	if(ak==&(moldyn->atom[DATOM])) {
		printf("force 3bp (k2): [%d %d %d]\n",ai->tag,aj->tag,ak->tag);
		printf("  adding %f %f %f\n",force.x,force.y,force.z);
		printf("  total k: %f %f %f\n",ak->f.x,ak->f.y,ak->f.z);
		printf("  angle: %f\n",acos(cos_theta)*360.0/(2*M_PI));
		printf("    d ij ik = %f %f\n",d_ij,d_ik);
	}
#endif

	/* virial */
	virial_calc(ai,&force,&dist_ik);
	
	/* force contribution to atom i */
	v3_scale(&force,&force,-1.0);
	v3_add(&(ai->f),&(ai->f),&force);

	/* increase k counter */
	exchange->kcnt+=1;

	return 0;
}

int albe_mult_i0_j3(t_moldyn *moldyn,t_atom *ai,t_atom *aj,u8 bc) {

	t_albe_mult_params *params;
	t_albe_exchange *exchange;

	params=moldyn->pot_params;
	exchange=&(params->exchange);

	/* increase j counter */
	exchange->j2cnt+=1;

	return 0;
}

int albe_mult_check_2b_bond(t_moldyn *moldyn,t_atom *itom,t_atom *jtom,u8 bc) {

	t_albe_mult_params *params;
	t_3dvec dist;
	double d;
	u8 brand;

	v3_sub(&dist,&(jtom->r),&(itom->r));
	if(bc) check_per_bound(moldyn,&dist);
	d=v3_absolute_square(&dist);

	params=moldyn->pot_params;
	brand=itom->brand;

	if(brand==jtom->brand) {
		if(d<=params->S2[brand])
			return TRUE;
	}
	else {
		if(d<=params->S2mixed)
			return TRUE;
	}

	return FALSE;
}
